﻿// COSC 4P76 - Assignment 1
// Michael Gircys (mg12vp, 5269550)
// 2016-03-01

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data;
using System.IO;

namespace KMeansClusterer
{
    public static class DataTableHelper
    {

        // Generates a datatable from a delimited string
        public static DataTable TableFromText( String content, Boolean IncludesHeader = false, Char delimiter = ',', Type colType = null )
        {
            var Rows = content.Replace("\r","").Split('\n');
            var Row0 = Rows[0].Split(delimiter);
            var RowCount = Rows.Length;
            var ColCount = Row0.Length;
            
            var dt = new DataTable();
            for(int x = 0; x < ColCount; x++)
            {
                dt.Columns.Add();
                dt.Columns[x].DataType   = colType ?? typeof(String);
                if(IncludesHeader)
                    dt.Columns[x].ColumnName = Row0[x];
            }
            for(int y = 0; y < RowCount; y++)
            {
                if(y==0 && IncludesHeader) continue;
                if(String.IsNullOrWhiteSpace(Rows[y])) continue;

                var dr = dt.NewRow();
                for(int x = 0; x < ColCount; x++)
                    dr[x] = Rows[y].Split(delimiter)[x];
                dt.Rows.Add(dr);                
            }

            return dt;
        }

        // Generates a datatable suitable for neural net training from an existing datatable
        // with expected format. 
        // It is assumed that all but the last column are attributes, and the possible 
        // classifications can be expressed as the set of distinct values in the last column.
        // The output will replace the classification column with multiple columns: one for
        // each possible classification, with a value of 1.0 where the sample holds the class
        // and 0.0 otherwise.
        // The ordered list of found classifications is returned through the output parameter.
        public static DataTable LastColumnClassification( DataTable dtIn, out List<String> classes )
        {
            // find distinct classifications
            classes = new List<String>();
            var classIdx = dtIn.Columns.Count-1;
            for(int r = 0; r < dtIn.Rows.Count; r++)
            {
                var curClass = dtIn.Rows[r][classIdx].ToString();
                if( !classes.Contains(curClass) )
                    classes.Add(curClass);
            }

            // copy to final table format
            var dt = new DataTable();
            for(int c = 0; c < classIdx; c++)
                dt.Columns.Add( new DataColumn( "Input-"+(c+1), typeof(double) ) );
            for(int c = 0; c < classes.Count; c++)
                dt.Columns.Add( new DataColumn( "Class-"+(c+1), typeof(double) ) );
            for(int r = 0; r < dtIn.Rows.Count; r++)
            {
                var dr = dt.NewRow();
                var curClassIdx = classes.IndexOf(dtIn.Rows[r][classIdx].ToString());
                for(int c = 0; c < classIdx; c++)
                    dr[c] = Double.Parse(dtIn.Rows[r][c].ToString());
                for(int c = 0; c < classes.Count; c++)
                    dr[classIdx + c] = ((c==curClassIdx) ? 1.0 : 0.0);

                dt.Rows.Add(dr);
            }

            return dt;
        }

        // Generates a datatable from a delimited text file
        public static DataTable Import(string filename, bool hasHeaderRow, char delim)
        {
            var inStr = File.ReadAllText(filename);
            var dt = TableFromText( inStr, hasHeaderRow, delim );
            return dt;
        }

        // Export datatable to comma-delimited text file
        public static String ToCSV( this DataTable dt )
        {
            var sb = new StringBuilder();

            for(int r = -1; r < dt.Rows.Count; r++)
            {
                for(int c = 0; c < dt.Columns.Count; c++)
                {
                    String val = String.Empty;
                    if(r < 0)
                            val = dt.Columns[c].ColumnName;
                    else    val = dt.Rows[r][c].ToString();

                    sb.Append( val );
                    if(c != dt.Columns.Count-1)
                        sb.Append(", ");
                }
                sb.AppendLine();
            }

            return sb.ToString();
        }

        // Append to the current datatable all rows from a provided data table
        // Assumes equivalent column definition
        public static void Union( this DataTable dt1, DataTable dt2 )
        {
            for(int r = 0; r < dt2.Rows.Count; r++)
                dt1.Rows.Add( dt2.Rows[r].ItemArray );
        }

        // Find simple aggregate statistics for all entries in a given column
        public static AggregateSet ColumnAggregates( this DataTable dt, int columnIndex )
        {
            var agg = new AggregateSet();
            foreach( DataRow row in dt.Rows )
                agg.Add( (double)(row[columnIndex]) );
            return agg;
        }

        // Why was this not already a thing?!
        public static IEnumerable<DataColumn> AsEnumerable( this DataColumnCollection cols )
        {
            return cols.OfType<DataColumn>();
        }

    }
}
