﻿// COSC 4P76 - Assignment 2
// Michael Gircys (mg12vp, 5269550)
// 2016-03-13

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace KMeansClusterer
{

    // Choices for distance functions
    public enum DistanceFunctionType
    {
        Euclidean,
        Euclidean_Square,
        Manhattan,
        Chebyshev,
        Canberra
    }


    // delegate signature for distance functions
    public delegate double DistanceFunctionDelegate(IEnumerable<double> pt1, IEnumerable<double> pt2);


    class DistanceFunctions
    {

        // Obtain distance function
        public static DistanceFunctionDelegate Get(DistanceFunctionType funcType)
        {
            switch( funcType )
            {
                case DistanceFunctionType.Euclidean:
                    return F_Euclidean;
                case DistanceFunctionType.Euclidean_Square:
                    return F_Euclidean_Square;
                case DistanceFunctionType.Manhattan:
                    return F_Manhattan;
                case DistanceFunctionType.Chebyshev:
                    return F_Chebyshev;
                case DistanceFunctionType.Canberra:
                    return F_Canberra;

                default:
                    return F_Euclidean_Square;
            }
        }


        // Fast / Square Euclidean Distance
        // (Minkowski 2-Norm distance)
        public static double F_Euclidean_Square(IEnumerable<double> pt1, IEnumerable<double> pt2)
        {
            if( pt1.Count() != pt2.Count() ) throw new Exception("Mismatched coordinate count");
            var sum = 0.0;
            for(int i = 0; i < pt1.Count(); i++)
                sum += Math.Pow( pt2.ElementAt(i)-pt1.ElementAt(i) , 2 );
            return sum;
        }

        // Euclidean distance
        public static double F_Euclidean(IEnumerable<double> pt1, IEnumerable<double> pt2)
        {
            return Math.Sqrt(F_Euclidean_Square(pt1,pt2));
        }

        // Manhattan distance
        // (Minkowski 1-Norm distance)
        public static double F_Manhattan(IEnumerable<double> pt1, IEnumerable<double> pt2)
        {
            if( pt1.Count() != pt2.Count() ) throw new Exception("Mismatched coordinate count");
            var sum = 0.0;
            for(int i = 0; i < pt1.Count(); i++)
                sum += Math.Abs(pt2.ElementAt(i)-pt1.ElementAt(i));
            return sum;
        }

        // Chebyshev distance
        // (Minkowski Infinity-Norm distance)
        public static double F_Chebyshev(IEnumerable<double> pt1, IEnumerable<double> pt2)
        {
            if( pt1.Count() != pt2.Count() ) throw new Exception("Mismatched coordinate count");
            var dist = 0.0;
            for(int i = 0; i < pt1.Count(); i++)
                dist = Math.Max( dist, Math.Abs(pt2.ElementAt(i)-pt1.ElementAt(i)) );
            return dist;
        }

        // Canberra distance
        // (Weighted Euclidean)
        public static double F_Canberra(IEnumerable<double> pt1, IEnumerable<double> pt2)
        {
            if( pt1.Count() != pt2.Count() ) throw new Exception("Mismatched coordinate count");
            var sum = 0.0;
            for(int i = 0; i < pt1.Count(); i++)
            {
                var v1 = pt1.ElementAt(i);
                var v2 = pt2.ElementAt(i);
                sum += Math.Abs(v2-v1) / ( Math.Abs(v2) + Math.Abs(v1) );
            }
            return sum;
        }


    }
}
