﻿// COSC 4P76 - Assignment 2
// Michael Gircys (mg12vp, 5269550)
// 2016-03-13

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Data;

namespace KMeansClusterer
{
    // Implementation of the K-Means (partition based) clustering algorithm
    public class KMeansClusterer
    {
        public  int K;
        public  int MaxIterations = Int32.MaxValue;
        private DistanceFunctionDelegate  DistanceFunc;
        private DistanceFunctionType     _DistanceType;
        public  DistanceFunctionType      DistanceType
        {
            get { return _DistanceType; }
            set { _DistanceType = value; DistanceFunc = DistanceFunctions.Get(value); }
        }
        

        public KMeansClusterer(){}
        public KMeansClusterer( int k, DistanceFunctionType distanceMetric )
        {
            K = k;
            DistanceType = distanceMetric;
        }


        public DataTable Cluster( DataTable points )
        {
            // The number of point dimensions is the number of columns of type double
            var pointDimens = points.Columns.AsEnumerable().Where( col => col.DataType == typeof(double) ).Count();

            // Setup cluster and centroid structures
            var clusters = points.Copy(); // structure and data.
            clusters.Columns.Add("ClusterGroup", typeof(int));
            clusters.Columns.Add("IsCentroid",   typeof(bool));
            foreach(DataRow point in clusters.Rows)
            {
                point["ClusterGroup"] = Int32.MaxValue;
                point["IsCentroid"] = false;
            }
            var centroids = clusters.Clone(); // structure only.

            // Determine data set extents.
            var aggs = new AggregateSet[pointDimens];
            for(int c = 0; c < aggs.Length; c++)
                aggs[c] = clusters.ColumnAggregates(c);

            // Initialze centroids
            for(int i = 0; i < K; i++)
            {
                //var centroid = centroids.NewRow();
                //centroid["ClusterGroup"] = i;
                //centroid["IsCentroid"] = true;
                //for(int c = 0; c < aggs.Length; c++)
                //    centroid[c] = RandomGenerator.Get().NextDouble( aggs[c].Min, aggs[c].Max );
                //centroids.Rows.Add(centroid);

                var randPoint = RandomGenerator.Get().Next(clusters.Rows.Count);
                centroids.Rows.Add( clusters.Rows[randPoint].ItemArray );
                centroids.Rows[i]["ClusterGroup"] = i;
                centroids.Rows[i]["IsCentroid"] = true;
            }

            var isStable = false;
            for( var gen = 0; !isStable && gen <= MaxIterations; gen++ )
            {   
                // Assign points to centroids
                isStable = (gen==0) ? false : true;
                foreach( DataRow point in clusters.Rows )
                {
                    var origCluster = (int)point["ClusterGroup"];
                    var minDist = Double.PositiveInfinity;
                    foreach( DataRow centroid in centroids.Rows )
                    {
                        var dist = DistanceFunc.Invoke( 
                                point   .ItemArray.Take(pointDimens).OfType<double>(),
                                centroid.ItemArray.Take(pointDimens).OfType<double>()
                            );
                        if( dist < minDist )
                        {
                            point["ClusterGroup"] = (int)centroid["ClusterGroup"];
                            minDist = dist;
                        }
                    }

                    // If the group changed, we're not stable, and should repeat.
                    if( origCluster != ((int)point["ClusterGroup"]) ) 
                        isStable = false;
                }

                // Maybe?
                // Don't adjust mean if we're happing with the grouping
                //if(isStable) continue;

                // Update centroid mean:
                // Initial new aggregate
                var centroidsAggs = new AggregateSet[K, pointDimens];
                foreach( DataRow centroid in centroids.Rows )
                for( int c = 0; c < pointDimens; c++)
                    centroidsAggs[ (int)centroid["ClusterGroup"], c] = new AggregateSet();
                // Find means
                foreach( DataRow point in clusters.Rows )
                for( int c = 0; c < pointDimens; c++)
                    centroidsAggs[ (int)point["ClusterGroup"], c].Add( (double)point[c] );
                // Update centroid
                foreach( DataRow centroid in centroids.Rows )
                for( int c = 0; c < pointDimens; c++)
                    if(centroidsAggs[(int)centroid["ClusterGroup"], c].Count > 0)
                        centroid[c] = centroidsAggs[(int)centroid["ClusterGroup"], c].Avg;

                if(gen == MaxIterations)
                    Console.WriteLine("Clustering stopped after reaching iteration limit.");
            }

            // Return clusters with centroids
            clusters.Union(centroids);
            return clusters;
        }
        

    }
}
