﻿// COSC 4P76 - Assignment 2
// Michael Gircys (mg12vp, 5269550)
// 2016-03-13

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Data;

namespace KMeansClusterer
{
    class Program
    {
        static String  DataFile;
        static String  LogFolder;
        static Boolean DataHasHeader;

        static int Runs = 1;
        static int K = 1;
        static int MaxIterations = 1000;
        static DistanceFunctionType DistanceFunc = DistanceFunctionType.Euclidean;
        static Boolean BatchRun = false;
        static Nullable<Int32> Seed = null;


        // Parse the input CLI arguments for neural net configuration
        static void GetConfig(string[] args)
        {
            var foundErrors = false;
            for(int p = 0; p < args.Length; p++)
            {
                try
                {
                    var param = args[p];

                    if(param.ToLower().Equals("-in"))
                    { DataFile = args[++p]; continue; }
                    if(param.ToLower().Equals("-out"))
                    { LogFolder = args[++p]; continue; }
                    if(param.ToLower().Equals("-inheader"))
                    { DataHasHeader = true; continue; }

                    if(param.ToLower().Equals("-r"))
                    { Runs = Int32.Parse(args[++p]); continue; }

                    if(param.ToLower().Equals("-k"))
                    { K = Int32.Parse(args[++p]); continue; }

                    if(param.ToLower().Equals("-max"))
                    { MaxIterations = Int32.Parse(args[++p]); continue; }

                    if(param.ToLower().Equals("-dist"))
                    {
                        DistanceFunc = (DistanceFunctionType)Enum.Parse( typeof(DistanceFunctionType), args[++p] );
                        continue;
                    }

                    if(param.ToLower().Equals("-seed"))
                    { Seed = Int32.Parse(args[++p]); continue; }

                    if(param.ToLower().Equals("-batch"))
                    { BatchRun = true; continue; }
                }
                catch( Exception ex )
                {
                    foundErrors = true;
                    Console.WriteLine(ex.Message);
                    Console.WriteLine(ex.StackTrace);
                }
            }
            if(foundErrors)
            {
                Console.WriteLine("Errors found. Continue [Y/N]?");
                var res = Console.ReadLine();
                if(!res.ToLower().StartsWith("Y"))
                    Environment.Exit(Environment.ExitCode);
            }
        }

        // Perform any further initializations based on the received configuration.
        static void Initialize()
        {
            if(LogFolder == null)
                LogFolder = Path.GetDirectoryName(DataFile);

            RandomGenerator.Initialize();
            if(Seed.HasValue)
                RandomGenerator.Initialize(Seed.Value);
        }

        static void Main(string[] args)
        {

            GetConfig(args);
            Initialize();

            if(!BatchRun)
            {
                Console.WriteLine("Press any key to begin...");
                Console.ReadKey();
            }

            // Import the specified data file.
            var dataStr = File.ReadAllText(DataFile);
            var dt = DataTableHelper.TableFromText( dataStr, DataHasHeader, ',', typeof(double) );

            // Initialize clusterer
            var clusterer = new KMeansClusterer( K, DistanceFunc );
            clusterer.MaxIterations = MaxIterations;

            // Perpare analytics
            var analyzer  = new ClusterAnalytics( DistanceFunc );
            var analytics = new DataTable();
            analytics.Columns.Add("Run",  typeof(int));
            analytics.Columns.Add("Dunn", typeof(double));

            for( int run = 0; run < Runs; run++ )
            {
                // Cluster!
                Console.WriteLine("Clustering.");
                var clusters = clusterer.Cluster(dt);
                
                // Sort and Export
                var dv = clusters.DefaultView;
                dv.Sort = "IsCentroid DESC, ClusterGroup ASC";
                File.WriteAllText( Path.Combine(LogFolder,"Run"+run.ToString()+".csv"), dv.ToTable().ToCSV());

                // Analyze
                Console.WriteLine("Analyzing.");
                var runAnalytics = analytics.NewRow();
                runAnalytics["Run"]  = run;
                runAnalytics["Dunn"] = analyzer.Dunn_Index( 
                        clusters, 
                        ClusterAnalytics.IntraDistanceType.Max, 
                        ClusterAnalytics.InterDistanceType.Closest
                    );
                analytics.Rows.Add(runAnalytics);
            }
            File.WriteAllText( Path.Combine(LogFolder,"Analytics.csv"), analytics.ToCSV());


            if(!BatchRun)
            {
                Console.WriteLine("Press any key to close...");
                Console.ReadKey();
            }
        }

    }
}
